//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Mon May 27 21:13:29 PDT 1996
// Last Modified: Mon May 27 21:13:38 PDT 1996
// Filename:      /user/c/craig/lang/c/drum/examples/eecho4.cc
// Syntax:        C++ (Gnu C++ compiler: /Local/Unix/bin/g++)
// $Smake:        g++ -O -o %b %f -I../include -L../lib %
//                -ldrum -lmidi -loptions && strip %b
//                Type "../bin/smake FILENAME" to compile this file.
//

#include "main.cc"

CharLoop major1;
CharLoop minor1;
CharLoop major2;
CharLoop minor2;

EventBuffer eventBuffer(2000);

void makeMajorLoop(CharLoop& loop, int direction = 1, int transp = 0) {
   int i;
   char keysu[8] = {0, 2, 4, 5, 7, 9, 11, 12};
   char keysd[8] = {12, 11, 9, 7, 5, 4, 2, 0};

   if (direction == 1) {
      for (i=0; i<8; i++) keysu[i] += transp;
      loop.setLoop(keysu, 8);
   } else {
      for (i=0; i<8; i++) keysd[i] += transp;
      loop.setLoop(keysd, 8);
   }
   loop.setPattern(CYCLIC);
}


void makeMinorLoop(CharLoop& loop, int direction = 1, int transp = 0) {
   int i;
   char keysu[8] = {0, 2, 3, 5, 7, 8, 11, 12};
   char keysd[8] = {12, 11, 8, 7, 5, 3, 2, 0};

   if (direction == 1) {
      for (i=0; i<8; i++) keysu[i] += transp;
      loop.setLoop(keysu, 8);
   } else {
      for (i=0; i<8; i++) keysd[i] += transp;
      loop.setLoop(keysd, 8);
   }
   loop.setPattern(CYCLIC);
}


/***************** Beginning of user functions ****************************/

//////////////////////////////
//
// echoFunction -- creates a note each time it is called according
//	to the velocity of note.
//

static void echoFunction(Note* note) {
   static Note tempnote;
   static int newVel;

   // make the echo note:
   tempnote.setStatus(NONE); 			// prevent extra note off
   tempnote.setOnDur(t_time, (int) (note->getOffTime() * drum.p2ps() * 2));
   tempnote.play(note->getKey(), note->getVel(), note->getChan());
   eventBuffer.insert(&tempnote);

   // update the velocity of the next echo:
   newVel = note->getVel() - drum.p1ps(3,20);
   if (newVel <= 0) { 
      note->off();
      return;
   } else {
      note->setVel(note->getVel() - drum.p1ps(3,20));
   }

   // set the time for the next note to be generated:
   note->setOnTime(note->getOnTime() + note->getOffTime());
}

//////////////////////////////
//
// upechoFunction -- rises from nothing then back down again
//

class UpechoParameters {
   public:
      int max;
      int direction;
      int vel;
      CharLoop* notes;
};

#define PARAMS ((UpechoParameters*)((FunctionNote*)note)->parameters())

static void upechoFunction(Note* note) {
   static Note tempnote;
   static int newVel;

   // make the echo note:
   tempnote.setStatus(NONE);                    // prevent extra note off
   tempnote.setOnDur(t_time, (int) (note->getOffTime() * drum.p2ps() * 2));
   tempnote.play(PARAMS->notes->next() + note->getKey(), 
                 note->getVel(), 
                 note->getChan());
   eventBuffer.insert(&tempnote);

   // update the velocity of the next echo:
   PARAMS->vel += PARAMS->direction * drum.p1ps(3,20);
   if (PARAMS->vel >= PARAMS->max) {
      PARAMS->vel = PARAMS->max;
      PARAMS->direction = -1;
   }
   if (PARAMS->vel <= 0) {
      delete PARAMS;
      note->off();
      return;
   } else {
      note->setVel(PARAMS->vel);
   }

   // set the time for the next note to be generated:
   note->setOnTime(note->getOnTime() + note->getOffTime());
}


/*---------------- Documentation functions -------------------------------*/

void description(void) {
   cout
   << "______________________________________________________________ \n"
   << "| Description:                                                |\n"
   << "|    Each baton triggers an echo algorithm.  The echo notes   |\n"
   << "| are created one at a time as they are needed.               |\n"
   << "|                           -- Craig Stuart Sapp, 27 May 1996 |\n"
   << "---------------------------------------------------------------\n"
   ;
}


void programhelp(void) {
   cout
   << ">                                                              \n"
   << "> Baton 1: Trigger echo                                        \n"
   << ">    x1t: Keyno                                                \n"
   << ">    y1t: Echo rate                                            \n"
   << ">    w1t: Initial key velocity                                 \n"
   << "> Baton 2:                                                     \n"
   << ">    x2t: Keyno                                                \n"
   << ">    y2t: Echo rate                                            \n"
   << ">    w2t: Initial key velocity                                 \n"
   << ">                                                              \n"
   << ">    p1p: Decay rate of the echos, global.                     \n"
   << ">    p2p: Sustain amount of echo notes, global.                \n"
   << ">    p3p: Channel of echo, local.                              \n"
   << ">   +b14: Command mode.                                        \n"
   << ">   +b15: This help screen.                                    \n"
   << ">                                                              \n"
   ;
}

/*---------------- Initialization functions ------------------------------*/

void initialization(void) { 
   description(); 
   eventBuffer.setPollRate(10);
   drum.setPollRate(40);
   makeMajorLoop(major1, -1);
   makeMinorLoop(minor1, -1);
   makeMajorLoop(major2, 1);
   makeMinorLoop(minor2, 1);
}

/*---------------- Main loop functions -----------------------------------*/

void mainloopalgorithms(void) { 
   eventBuffer.checkPoll();
   drum.checkPollPots();
}

/*---------------- Trigger response functions ----------------------------*/

void upecho(int pitch, int maxvelocity, int separation, int channel) {
   UpechoParameters* params;
   params = new UpechoParameters;
   params->max = maxvelocity;
   params->vel = 1;
   params->direction = 1;
   if (maxvelocity < 100)
      params->notes = &major2;
   else 
      params->notes = &minor2;
   static FunctionNote tempnote;
   tempnote.setChan(channel);
   tempnote.setKeyno(pitch);
   tempnote.setVel(1);
   tempnote.setOffTime(separation);
   tempnote.setOnTime(t_time);
   tempnote.setFunction(&upechoFunction);
   tempnote.setParameters(params);
   eventBuffer.insert(&tempnote);
}

void downecho(int pitch, int maxvelocity, int separation, int channel) {
   UpechoParameters* params;
   params = new UpechoParameters;
   params->max = maxvelocity;
   params->vel = 1;
   params->direction = 1;
   if (maxvelocity < 100)
      params->notes = &major1;
   else 
      params->notes = &minor1;
   static FunctionNote tempnote;
   tempnote.setChan(channel);
   tempnote.setKeyno(pitch);
   tempnote.setVel(1);
   tempnote.setOffTime(separation);
   tempnote.setOnTime(t_time);
   tempnote.setFunction(&upechoFunction);
   tempnote.setParameters(params);
   eventBuffer.insert(&tempnote);
}

void stick1trig(void) { 
   cout << "Baton 1 triggered\n"; 
   downecho(drum.x1ts(A0, C7), drum.w1ts(50, 127), drum.y1ts(20, 1000),
        drum.p3ps(CH_1, CH_16));
}

void stick2trig(void) { 
   cout << "Baton 2 triggered\n"; 
   upecho(drum.x2ts(A0, C7), drum.w2ts(50, 127), drum.y2ts(20, 1000),
        drum.p3ps(CH_1, CH_16));
}

void b14plustrig(void) { cout << "B14+ triggered\n"; commandmode(); }
void b15plustrig(void) { cout << "B15+ triggered\n"; programhelp(); }

/*---------------- Poll response functions -------------------------------*/

void stick1pollresponse(void) { }
void stick2pollresponse(void) { }
void potspollresponse(void) { }

/*---------------- Actions to take just before leaving program -----------*/

void cleanup(void) { }

/***************** End of user functions **********************************/

