// hotcold.cc -- The game of hot and cold for the radio drum.
//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Tue Mar 19 13:35:34 PST 1996
// Last Modified: Mon Apr  8 21:35:04 PDT 1996
// Filename:      /user/c/craig/lang/c/drum/examples/hotcold.cc
// Syntax:        C++ (Gnu C++ compiler: /Local/Unix/bin/g++)
// $Smake:        cc -O -o %b %f -I../include -L../lib -lg++ %
//                -ldrum -lmidi -loptions && strip %b
//                Type "../bin/smake FILENAME" to compile this file.
//

#include <math.h>
#include "main.cc"
#include "psr-520.h"


/***************** Beginning of user function *****************************/

Note  xVoice(CH_1, PSR_ACOUSTIC_GRAND_PIANO); // voice for x-axis
Note  yVoice(CH_2, PSR_MUSIC_BOX);            // voice for y-axis
Note  zVoice(CH_3, PSR_MARIMBA);              // voice for z-axis

int  xrand, yrand, zrand;           // random point in baton space
int  status = OFF;                  // test whether to play notes
long timer;                         // time how long it takes to find point


/*---------------- Documentation functions -------------------------------*/

void description(void) {
   cout
   << "______________________________________________________________ \n"
   << "| Description:                                                |\n"
   << "|    Hit the drum with a baton to start a search for a random |\n"
   << "| point.  Notes stop playing when the 3-D point is found.     |\n"
   << "| You are hot when the pitches rise; cold when pitch drops.   |\n"
   << "| Each axis has a voice associated with itself.               |\n"
   << "| Loudness of notes is determined by Baton 2 position.        |\n"
   << "| Tempo of notes is determined by Baton 2's y-position.       |\n"
   << "|                        -- Craig Stuart Sapp, 21 April 1996  |\n"
   << "---------------------------------------------------------------\n"
   ;
}


void programhelp(void) {
   cout
   << ">                                                              \n"
   << "> Baton 1:                                                     \n"
   << ">   trig: Start the hot/cold game.                             \n"
   << ">    x1p: X-axis note -- higher if warmer, lower if colder.    \n"
   << ">    y1p: Y-axis note -- higher if warmer, lower if colder.    \n"
   << ">    z1p: Z-axis note -- higher if warmer, lower if colder.    \n"
   << "> Baton 2:                                                     \n"
   << ">    x2p: Attack velocity of x direction voice.                \n"
   << ">    y2p: Attack velocity of y direction voice and tempo.      \n"
   << ">    z2p: Attack velocity of z direction voice.                \n"
   << ">                                                              \n"
   << ">    p1p: Turns hot/cold game on or off.                       \n"
   << ">    p2p: Timbre of the x-axis.                                \n"
   << ">    p3p: Timbre of the z-axis.                                \n"
   << ">   +b14: Command mode.                                        \n"
   << ">   +b15: This help screen.                                    \n"
   << ">                                                              \n"
   ;
}


/*---------------- Initialization algorithms -----------------------------*/

void initialization(void) { 
   drum.setPollRate(200);       // poll rate of baton set to 200 milliseconds
   description();
   xVoice.sendTimbre();
   yVoice.sendTimbre();
   zVoice.sendTimbre();
}


/*---------------- Main loop algorithms ----------------------------------*/

void mainloopalgorithms(void) { 
   if (status == ON && drum.checkPollTime()) {

      // play the notes for each axis if a new note space is entered:
      if (drum.cx1p(A0, C7)) {
         xVoice.play(C7-abs(xrand-drum.x1ps(A0, C7)), drum.x2p);
         cout << " x = "   << C7-abs(xrand-drum.x1ps(A0, C7));
      } 
      if (drum.cy1p(A0, C7))  {
         yVoice.play(C7-abs(yrand-drum.y1ps(A0, C7)), drum.y2p);
         cout << " y = "   << C7-abs(yrand-drum.y1ps(A0, C7));
      }
      if (drum.cz1p(A0, C7))  {
         zVoice.play(C7-abs(zrand-drum.z1ps(A0, C7)), 127 - drum.z2p);
         cout << " z = "   << C7-abs(zrand-drum.z1ps(A0, C7));
      }

      cout << endl;


      // if the 3-D point was found, turn off the notes and print time:
      if (   abs(drum.x1ps(A0, C7) - xrand) < 10  &&
             abs(drum.y1ps(A0, C7) - yrand) < 10  &&
             abs(drum.z1ps(A0, C7) - zrand) < 10       ) {
         status = OFF;
         xVoice.off();
         yVoice.off();
         zVoice.off();
         cout << "Time to find point: " << (t_time-timer)/1000.0 << " sec.\n";
      }
      drum.pollAll();
   }
}


/*---------------- Trigger response functions ----------------------------*/

void stick1trig(void) { 
   // if not already searching for a point, choose a new point and start timer:
   if (status == OFF) {
      status = ON;
      xrand = rand() % (C7-A0) + A0;
      yrand = rand() % (C7-A0) + A0;
      zrand = (rand() % 44) + 44;        // limit z-axis close to drum surface
      timer = t_time;
      cout << "random point chosen\n"; 
   }
}

void stick2trig(void) { }

void b14plustrig(void) { 
   status = OFF;
   xVoice.off();
   yVoice.off();
   zVoice.off();
   commandmode(); 
}

void b15plustrig(void) { programhelp(); }

/*---------------- Poll response algorithms ------------------------------*/

void stick1pollresponse(void) { }
void stick2pollresponse(void) { 
   drum.setPollRate(80 + 3 * drum.y2p); 
}

void potspollresponse(void) { 
   if (status == ON && drum.p1ps() < 0.1) {
      status = OFF;
      xVoice.off();
      yVoice.off();
      zVoice.off();
   }

  if (drum.p2ps(0,63) != drum.p2pos(0,63)) {
      midi.pc(CH_1, drum.p2ps(0,63));
      cout << "\t\t\t\tTimbre X = " << drum.p2ps(0, 63) << endl;
  }
  if (drum.p3ps(0,63) != drum.p3pos(0,63)) {
      midi.pc(CH_3, drum.p3ps(0,63));
      cout << "\t\tTimbre Z = " << drum.p3ps(0, 63) << endl;
  }

}

/*---------------- Actions to take just before leaving program -----------*/

void cleanup(void) { }

/***************** End of user functions **********************************/

