//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Wed Jan 21 22:46:30 GMT-0800 1998
// Last Modified: Sat Jan  9 06:05:24 PST 1999
// Filename:      ...sig/code/control/MidiInPort/linux/MidiInPort_linux.cpp
// Syntax:        C++ 
// $Smake:        g++ -g -c %b.cpp -I../../../../include && rm -f %b.o
//

#ifdef LINUX

#include "MidiInPort_linux.h"
#include <iostream.h>
#include <stdlib.h>
#include <pthread.h>
#include <linux/soundcard.h>


#define DEFAULT_INPUT_BUFFER_SIZE (1024)

// initialized static variables
int       MidiInPort_linux::numDevices                     = 0;
int       MidiInPort_linux::objectCount                    = 0;
int*      MidiInPort_linux::portObjectCount                = NULL;
CircularBuffer<MidiMessage>** MidiInPort_linux::midiBuffer = NULL;
int       MidiInPort_linux::channelOffset                  = 0;
SigTimer  MidiInPort_linux::midiTimer;
int*      MidiInPort_linux::pauseQ                         = NULL;
pthread_t MidiInPort_linux::midiInThread;    


//////////////////////////////
// 
// MidiInPort_linux::MidiInPort_linux
//	default values: autoOpen = 1
//

MidiInPort_linux::MidiInPort_linux(void) {
   if (objectCount == 0) {
      initialize();
   }
   objectCount++;

   port = -1;
   setPort(0);
}


MidiInPort_linux::MidiInPort_linux(int aPort, int autoOpen) {
   if (objectCount == 0) {
      initialize();
   }
   objectCount++;

   port = -1;
   setPort(aPort);
   if (autoOpen) {
      open();
   }
}



//////////////////////////////
//
// MidiInPort_linux::~MidiInPort_linux
//

MidiInPort_linux::~MidiInPort_linux() {
   objectCount--;
   if (objectCount == 0) {
      deinitialize();
   } else if (objectCount < 0) {
      cerr << "Error: bad MidiInPort_linux object count!: " 
           << objectCount << endl;
      exit(1);
   }
}



//////////////////////////////
//
// MidiInPort_linux::close
//

void MidiInPort_linux::close(void) {
   pauseQ[getPort()] = 1;
}



//////////////////////////////
//
// MidiInPort_linux::closeAll --
//

void MidiInPort_linux::closeAll(void) {
   Sequencer_linux::close();
}



//////////////////////////////
//
// MidiInPort_linux::extract -- returns the next MIDI message
//	received since that last extracted message.
//

MidiMessage MidiInPort_linux::extract(void) {
   return midiBuffer[getPort()]->extract();
}



//////////////////////////////
//
// MidiInPort_linux::getBufferSize -- returns the maximum possible number
//	of MIDI messages that can be stored in the buffer
//

int MidiInPort_linux::getBufferSize(void) {
   return midiBuffer[getPort()]->getSize();
}



//////////////////////////////
//
// MidiInPort_linux::getChannelOffset -- returns zero if MIDI channel 
//     offset is 0, or 1 if offset is 1.
//

int MidiInPort_linux::getChannelOffset(void) const {
   return channelOffset;
}



//////////////////////////////
//
// MidiInPort_linux::getCount -- returns the number of unexamined
//	MIDI messages waiting in the input buffer.
//

int MidiInPort_linux::getCount(void) {
   return midiBuffer[getPort()]->getCount();
}



//////////////////////////////
//
// MidiInPort_linux::getName -- returns the name of the port.
//	returns "" if no name. Name is valid until all instances
//      of MIDI classes are.
//

const char* MidiInPort_linux::getName(void) {
   return getInputName(getPort());
}


const char* MidiInPort_linux::getName(int i) {
   return getInputName(i);
}



//////////////////////////////
//
// MidiInPort_linux::getNumPorts -- returns the number of available
// 	ports for MIDI input
//

int MidiInPort_linux::getNumPorts(void) {
   return getNumInputs();
}



//////////////////////////////
//
// MidiInPort_linux::getPort -- returns the port to which this
//	object belongs (as set with the setPort function).
//

int MidiInPort_linux::getPort(void) {
   return port;
}



//////////////////////////////
//
// MidiInPort_linux::getPortStatus -- 0 if closed, 1 if open
//

int MidiInPort_linux::getPortStatus(void) {
   return is_open();
}



//////////////////////////////
//
// MidiInPort_linux::getTrace -- returns true if trace is on or false
//	if trace is off.  if trace is on, then prints to standard
// 	output the Midi message received.
//

int MidiInPort_linux::getTrace(void) {
   return trace;
}



//////////////////////////////
//
// MidiInPort_linux::insert
//

void MidiInPort_linux::insert(const MidiMessage& aMessage) {
   midiBuffer[getPort()]->insert(aMessage);
}



//////////////////////////////
//
// MidiInPort_linux::message
//

MidiMessage& MidiInPort_linux::message(int index) {
   CircularBuffer<MidiMessage>& temp = *midiBuffer[getPort()];
   return temp[index];
}



//////////////////////////////
//
// MidiInPort_linux::open -- returns true if MIDI input port was
//	opened.
//

int MidiInPort_linux::open(void) {
   return Sequencer_linux::open();
   pauseQ[getPort()] = 0;
}



//////////////////////////////
//
// MidiInPort_linux::pause -- stop the Midi input port from
//	inserting MIDI messages into the buffer, but keeps the
//	port open.  Use unpause() to reverse the effect of pause().
//

void MidiInPort_linux::pause(void) {
   pauseQ[getPort()] = 1;
}



//////////////////////////////
//
// MidiInPort_linux::setBufferSize -- sets the allocation
//	size of the MIDI input buffer.
//

void MidiInPort_linux::setBufferSize(int aSize) {
   midiBuffer[getPort()]->setSize(aSize);
}



//////////////////////////////
//
// MidiInPort_linux::setChannelOffset -- sets the MIDI chan offset, 
//     either 0 or 1.
//

void MidiInPort_linux::setChannelOffset(int anOffset) {
   switch (anOffset) {
      case 0:   channelOffset = 0;   break;
      case 1:   channelOffset = 1;   break;
      default:
         cout << "Error:  Channel offset can be only 0 or 1." << endl;
         exit(1);
   }
}



//////////////////////////////
//
// MidiInPort_linux::setPort
//

void MidiInPort_linux::setPort(int aPort) {
   if (aPort < 0 || aPort >= getNumPorts()) {
      cerr << "Error: maximum port number is: " << getNumPorts()-1
           << ", but you tried to access port: " << aPort << endl;
      exit(1);
   }

   if (port != -1) {
      portObjectCount[port]--;
   }
   port = aPort;
   portObjectCount[port]++;
}



//////////////////////////////
//
// MidiInPort_linux::setTrace -- if false, then don't print MIDI messages
// 	to the screen.
//

int MidiInPort_linux::setTrace(int aState) {
   int oldtrace = trace;
   if (aState == 0) {
      trace = 0;
   } else {
      trace = 1;
   }
   return oldtrace;
}



//////////////////////////////
//
// MidiInPort_linux::toggleTrace -- switches the state of trace
//	Returns the previous value of the trace variable.
//

void MidiInPort_linux::toggleTrace(void) {
   trace = !trace;
}
   


//////////////////////////////
//
// MidiInPort_linux::unpause -- enables the Midi input port 
//	to inserting MIDI messages into the buffer after the 
//	port is already open.
//

void MidiInPort_linux::unpause(void) {
   pauseQ[getPort()] = 0;
}



///////////////////////////////////////////////////////////////////////////
//
// Private functions
//



//////////////////////////////
//
// MidiInPort_linux::deinitialize -- sets up storage if necessary
//	This function should be called if the current object is
//	the first object to be created.
//

void MidiInPort_linux::deinitialize(void) {
   closeAll();
   if (midiBuffer != NULL) {
      for (int i=0; i<getNumPorts(); i++) {
/*  These lines cause a segmentation fault in linux:
         if (midiBuffer[i] != NULL)   delete [] midiBuffer[i];
         midiBuffer[i] = NULL;
*/
      }
      delete [] midiBuffer;
   }
   midiBuffer = NULL;
   if (portObjectCount != NULL)   delete [] portObjectCount;
   portObjectCount = NULL;
   if (pauseQ != NULL)   delete [] pauseQ;
   pauseQ = NULL;
}



//////////////////////////////
//
// MidiInPort_linux::initialize -- sets up storage if necessary
//	This function should be called if the current object is
//	the first object to be created.
//

void MidiInPort_linux::initialize(void) {
   // get the number of ports
   numDevices = Sequencer_linux::indevcount;
   if  (getNumPorts() <= 0) {
      cerr << "Error: no MIDI input devices" << endl;
      exit(1);
   }

   // allocate space for pauseQ, the port pause status
   if (pauseQ != NULL) delete [] pauseQ;
   pauseQ = new int[numDevices];

   // allocate space for object count on each port:
   if (portObjectCount != NULL) delete [] portObjectCount;
   portObjectCount = new int[numDevices];

   // allocate space for the Midi input buffers
   if (midiBuffer != NULL) delete [] midiBuffer;
   midiBuffer = new CircularBuffer<MidiMessage>*[numDevices];

   // initialize the static arrays
   for (int i=0; i<getNumPorts(); i++) {
      portObjectCount[i] = 0;
      pauseQ[i] = 0;
      midiBuffer[i] = new CircularBuffer<MidiMessage>;
      midiBuffer[i]->setSize(DEFAULT_INPUT_BUFFER_SIZE);
   }

   int flag = pthread_create(&midiInThread, NULL, 
      interpretMidiInputStreamPrivate, NULL);
   if (flag == -1) {
      cout << "Unable to create MIDI input thread." << endl;
      exit(1);
   }
   
}



///////////////////////////////////////////////////////////////////////////
//
// friendly functions 
//


//////////////////////////////
//
// interpretMidiInputStreamPrivate -- handles the MIDI input stream
//     for the various input devices.
//

void *interpretMidiInputStreamPrivate(void *) {

   int* argsExpected = NULL;     // MIDI parameter bytes expected to follow
   int* argsLeft     = NULL;     // MIDI parameter bytes left to wait for
   uchar packet[4];              // bytes for sequencer driver
   MidiMessage* message = NULL;  // holder for current MIDI message
   int newSigTime = 0;           // for millisecond timer
   int lastSigTime = -1;         // for millisecond timer
   int flag;                     // error flag thingy
   int zeroSigTime = -1;         // for timing incoming events
   int device = -1;              // for sorting out the bytes by input device

   static int count = 0;
   if (count != 0) {
      cerr << "Cannot run this function more than once" << endl;
      exit(1);
   } else {
      // allocate space for MIDI messages, each device has a different message
      // holding spot in case the messages overlap in the input stream
      message      = new MidiMessage[MidiInPort_linux::numDevices];
      argsExpected = new int[MidiInPort_linux::numDevices];
      argsLeft     = new int[MidiInPort_linux::numDevices];

      count++;
   }
   
   // interpret MIDI bytes as they come into the computer
   // and repackage them as MIDI messages.
   while (1) {
      ::read(MidiInPort_linux::sequencer_fd, &packet, sizeof(packet));

/*
      cout << "packet bytes "
           << " 0x" << hex << (int)packet[0]   // packet type
           << " 0x" << hex << (int)packet[1]   // possible MIDI byte
           << " 0x" << hex << (int)packet[2]   // device number
           << " 0x" << hex << (int)packet[3]   // unused for MIDI bytes
           << endl;
*/

      switch (packet[0]) {
         case SEQ_WAIT:
            // MIDI clock ticks ... the first MIDI message deltaTime is
            // calculated wrt the start of the MIDI clock.
            if (zeroSigTime < 0) {
               zeroSigTime = MidiInPort_linux::midiTimer.getTime();
            }
/* 
            int newTime;
            newTime = packet[3];
            newTime = (newTime << 8) | packet[2];
            newTime = (newTime << 8) | packet[1];
*/
            break;

         case SEQ_ECHO:
            // echo events
            break;

         case SEQ_MIDIPUTC:          // SEQ_MIDIPUTC = 5
            // determination of a full MIDI message from the input MIDI
            // stream is based here on the observation that MIDI status
            // bytes and subsequent data bytes are NOT returned in the same
            // read() call.  Rather, they are spread out over multiple read()
            // returns, with only a single value per return.  So if we
            // find a status byte, we then determine the number of expected
            // operands and process that number of subsequent read()s to
            // to determine the complete midi message.

/*
            cout << "byte: " << (int)packet[1] << endl;
*/

            device = packet[2];
           
            // ignore the active 0xfe and 0xf8 commands:
            if (packet[1] == 0xf8 || packet[1] == 0xfe) {
               break;
            }

            if (packet[1] & 0x80) {   // a command byte
               switch (packet[1] & 0xf0) {
                  case 0xf0:   argsExpected[device] = 0;    break;  // 0xf0 cmds not ready
                  case 0xc0:   argsExpected[device] = 1;    break;
                  case 0xd0:   argsExpected[device] = 1;    break;
                  default:     argsExpected[device] = 2;    break;
               }
               argsLeft[device] = argsExpected[device];

               newSigTime = MidiInPort_linux::midiTimer.getTime();
               message[device].time = newSigTime - zeroSigTime;

               message[device].p0() = packet[1];
               message[device].p1() = 0;
               message[device].p2() = 0;
               message[device].p3() = 0;
            } else if (argsLeft[device]) {
               if (message[device].time == 0) {
                  newSigTime = MidiInPort_linux::midiTimer.getTime();
                  message[device].time = newSigTime - zeroSigTime;
               }
                  
               if (argsLeft[device] == argsExpected[device]) {
                  message[device].p1() = packet[1];
               } else {
                  message[device].p2() = packet[1];
               }
               argsLeft[device]--;

               // if MIDI message is complete, setup for running status, and 
               // insert note into proper buffer.

               if (!argsLeft[device]) {
                  switch (message[device].p0() & 0xf0) {
                     case 0xc0:      argsLeft[device] = 1;      break;
                     case 0xa0:      argsLeft[device] = 1;      break;
                     default:        argsLeft[device] = 2;      break;
                  }

/*   
                  if (absolute) {
                     message[device].time = newSigTime - zeroSigTime;
                     lastSigTime = newSigTime;
                  } else {
                     if (lastSigTime == -1) {
                        message[device].time = 0;
                     } else {
                        message[device].time = newSigTime - lastSigTime;
                     }
                     lastSigTime = newSigTime;
                  }
*/
                  lastSigTime = newSigTime;

                  // insert the MIDI message into the appropriate buffer
                  // do not insert into buffer if the MIDI input device
                  // is paused (which can mean closed).  Or if the
                  // pauseQ array is pointing to NULL (which probably means that
                  // things are about to shut down).
                  if (MidiInPort_linux::pauseQ != NULL &&
                        MidiInPort_linux::pauseQ[device] == 0) {
                     MidiInPort_linux::midiBuffer[device]->insert(message[device]);
                     message[device].time = 0;
                  }
               }
            }
            break;

         default:
            break;
      } // end switch
   } // end while (1)

}
                                       


#endif  // LINUX



