//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Fri Dec 18 19:22:20 PST 1998
// Last Modified: Fri Jan  8 04:26:16 PST 1999
// Filename:      ...sig/code/control/MidiOutPort/linux/MidiOutPort_linux.cpp
// Syntax:        C++ 
// $Smake:        g++ -DLINUX -O3 -g -c %b.cpp -I../../../../include %
//                && rm %b.o
//


#ifdef LINUX

#include <iostream.h>
#include "MidiOutPort_linux.h"

// initialized static variables
int       MidiOutPort_linux::numDevices      = 0;
int       MidiOutPort_linux::objectCount     = 0;
int*      MidiOutPort_linux::portObjectCount = NULL;
int       MidiOutPort_linux::channelOffset   = 0;



//////////////////////////////
// 
// MidiOutPort_linux::MidiOutPort_linux
//	default values: autoOpen = 1
//


MidiOutPort_linux::MidiOutPort_linux(void) {
   if (objectCount == 0) {
      initialize();
   }
   objectCount++;

   port = -1;
   setPort(0);
}


MidiOutPort_linux::MidiOutPort_linux(int aPort, int autoOpen) {
   if (objectCount == 0) {
      initialize();
   }
   objectCount++;

   port = -1;
   setPort(aPort);
   if (autoOpen) {
      open();
   }
}



//////////////////////////////
//
// MidiOutPort_linux::~MidiOutPort_linux
//

MidiOutPort_linux::~MidiOutPort_linux() {
   objectCount--;
   if (objectCount == 0) {
      deinitialize();
   } else if (objectCount < 0) {
      cerr << "Error: bad MidiOutPort object count!: " << objectCount << endl;
      exit(1);
   }
}



//////////////////////////////
//
// MidiOutPort_linux::close
//

void MidiOutPort_linux::close(void) {
   // don't close anything, because the 
   // Linux driver keeps all of the ports open while the
   // midi driver (/dev/sequencer) is running.
}



//////////////////////////////
//
// MidiOutPort_linux::closeAll
//

void MidiOutPort_linux::closeAll(void) {
   // the Linux MIDI driver will close the /dev/sequencer device
   // which will close all MIDI output ports at the same time.
   Sequencer_linux::close();
}



//////////////////////////////
//
// MidiOutPort_linux::getChannelOffset -- returns zero if MIDI channel 
//     offset is 0, or 1 if offset is 1.
//

int MidiOutPort_linux::getChannelOffset(void) const {
   return channelOffset;
}



//////////////////////////////
//
// MidiOutPort_linux::getName -- returns the name of the port.
//	returns "" if no name. Name is valid until getName is called again.
//

const char* MidiOutPort_linux::getName(void) {
   return getOutputName(getPort());
}


const char* MidiOutPort_linux::getName(int i) {
   return Sequencer_linux::getOutputName(i);
}



//////////////////////////////
//
// MidiOutPort_linux::getNumPorts -- returns the number of available
// 	ports for MIDI output
//

int MidiOutPort_linux::getNumPorts(void) {
   return Sequencer_linux::getNumOutputs();
}



//////////////////////////////
//
// MidiOutPort_linux::getPort -- returns the port to which this
//	object belongs (as set with the setPort function).
//

int MidiOutPort_linux::getPort(void) {
   return port;
}



//////////////////////////////
//
// MidiOutPort_linux::getPortStatus -- 0 if closed, 1 if open
//

int MidiOutPort_linux::getPortStatus(void) {
   // Linux MIDI devices are all open at the same time,
   // so if one is open, then they all are.
   return is_open();
}



//////////////////////////////
//
// MidiOutPort_linux::getTrace -- returns true if trace is on or
//	false if off.  If trace is on, then prints to standard output
//	the Midi message being sent.
//

int MidiOutPort_linux::getTrace(void) {
   return trace;
}



//////////////////////////////
//
// MidiOutPort_linux::rawsend -- send the Midi command and its parameters
//

int MidiOutPort_linux::rawsend(int command, int p1, int p2) {
   write(getPort(), command);   
   write(getPort(), p1);   
   write(getPort(), p2);   

   return 1;   // can't tell if the send was successful, so assume that it was
}


int MidiOutPort_linux::rawsend(int command, int p1) {
   write(getPort(), command);   
   write(getPort(), p1);   
  
   return 1;   // can't tell if the send was successful, so assume that it was
}


int MidiOutPort_linux::rawsend(int command) {
   write(getPort(), command);

   return 1;   // can't tell if the send was successful, so assume that it was
}


int MidiOutPort_linux::rawsend(uchar* array, int size) {
   write(getPort(), array, size);
   
   return 1;   // can't tell if the send was successful, so assume that it was
}



//////////////////////////////
//
// MidiOutPort_linux::open -- returns true if MIDI output port was
//	opened.
//

int MidiOutPort_linux::open(void) {
   return Sequencer_linux::open();
}



//////////////////////////////
//
// MidiOutPort_linux::setChannelOffset -- sets the MIDI channel offset, 
//     either 0 or 1.
//

void MidiOutPort_linux::setChannelOffset(int anOffset) {
   switch (anOffset) {
      case 0:   channelOffset = 0;   break;
      case 1:   channelOffset = 1;   break;
      default:
         cout << "Error:  Channel offset can be only 0 or 1." << endl;
         exit(1);
   }
}



//////////////////////////////
//
// MidiOutPort_linux::setPort
//

void MidiOutPort_linux::setPort(int aPort) {
   if (aPort < 0 || aPort >= getNumPorts()) {
      cerr << "Error: maximum port number is: " << getNumPorts()-1
           << ", but you tried to access port: " << aPort << endl;
      exit(1);
   }

   if (port != -1) {
      portObjectCount[port]--;
   }
   port = aPort;
   portObjectCount[port]++;
}



//////////////////////////////
//
// MidiOutPort_linux::setTrace -- if false, then won't print
//      Midi messages to standard output.
//

int MidiOutPort_linux::setTrace(int aState) {
   int oldtrace = trace;
   if (aState == 0) {
      trace = 0;
   } else {
      trace = 1;
   }
   return oldtrace;
}



//////////////////////////////
//
// MidiOutPort_linux::toggleTrace
//

void MidiOutPort_linux::toggleTrace(void) {
   trace = !trace;
}



///////////////////////////////////////////////////////////////////////////
//
// Private functions
//



//////////////////////////////
//
// MidiOutPort_linux::deinitialize -- sets up storage if necessary
//	This function should be called if the current object is
//	the first object to be created.
//

void MidiOutPort_linux::deinitialize(void) {
   closeAll();
   if (portObjectCount != NULL) delete [] portObjectCount;
   portObjectCount = NULL;
}



//////////////////////////////
//
// MidiOutPort_linux::initialize -- sets up storage if necessary
//	This function should be called if the current object is
//	the first object to be created.
//

void MidiOutPort_linux::initialize(void) {
   // get the number of ports
   numDevices = getNumOutputs();
   if  (getNumPorts() <= 0) {
      cerr << "Error: no MIDI output devices" << endl;
      exit(1);
   }

   // allocate space for object count on each port:
   if (portObjectCount != NULL) delete [] portObjectCount;
   portObjectCount = new int[numDevices];


   // initialize the static arrays
   for (int i=0; i<getNumPorts(); i++) {
      portObjectCount[i] = 0;
   }
}



//////////////////////////////
//
// MidiOutPort_linux::setPortStatus
//

void MidiOutPort_linux::setPortStatus(int aStatus) {
   // not used in Linux implementation
}


#endif  // LINUX




